use serde::{de::Error, Deserialize, Serialize};

use crate::{
    constants::CMD_NAME,
    device_prober::get_xr_usb_devices,
    file_utils::get_writer,
    paths::get_config_dir,
    profile::Profile,
    profiles::{
        lighthouse::lighthouse_profile, openhmd::openhmd_profile, simulated::simulated_profile,
        survive::survive_profile, wivrn::wivrn_profile, wmr::wmr_profile,
    },
};
use std::{
    fs::File,
    io::BufReader,
    path::{Path, PathBuf},
};

fn default_win_size() -> [i32; 2] {
    [360, 400]
}

#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
pub struct Config {
    pub selected_profile_uuid: String,
    pub debug_view_enabled: bool,
    pub user_profiles: Vec<Profile>,
    #[serde(default = "default_win_size")]
    pub win_size: [i32; 2],
}

impl Default for Config {
    fn default() -> Self {
        Config {
            // TODO: handle first start with no profile selected
            selected_profile_uuid: "".to_string(),
            debug_view_enabled: false,
            user_profiles: vec![],
            win_size: default_win_size(),
        }
    }
}

impl Config {
    pub fn get_selected_profile(&self, profiles: &[Profile]) -> Profile {
        let def = || profiles.first().expect("No profiles found").clone();

        match profiles
            .iter()
            .find(|p| p.uuid == self.selected_profile_uuid)
        {
            Some(p) => p.clone(),
            None => match get_xr_usb_devices().first() {
                Some(dev) => match dev.get_default_profile() {
                    Some(p) => p,
                    None => def(),
                },
                None => def(),
            },
        }
    }

    pub fn config_file_path() -> PathBuf {
        get_config_dir().join(format!("{CMD_NAME}.json"))
    }

    fn from_path(path: &Path) -> Self {
        match File::open(path) {
            Ok(file) => {
                let reader = BufReader::new(file);
                match serde_json::from_reader(reader) {
                    Ok(config) => config,
                    Err(_) => Self::default(),
                }
            }
            Err(_) => Self::default(),
        }
    }

    fn save_to_path(&self, path: &Path) -> Result<(), serde_json::Error> {
        let writer = get_writer(path).map_err(serde_json::Error::custom)?;
        serde_json::to_writer_pretty(writer, self)
    }

    pub fn save(&self) {
        self.save_to_path(&Self::config_file_path())
            .expect("Failed to save config");
    }

    pub fn get_config() -> Self {
        Self::from_path(&Self::config_file_path())
    }

    pub fn set_profiles(&mut self, profiles: &[Profile]) {
        self.user_profiles = profiles.iter().filter(|p| p.editable).cloned().collect();
    }

    pub fn profiles(&self) -> Vec<Profile> {
        let mut profiles = vec![
            lighthouse_profile(),
            survive_profile(),
            wivrn_profile(),
            wmr_profile(),
            openhmd_profile(),
            simulated_profile(),
        ];
        profiles.extend(self.user_profiles.clone());
        profiles.sort_unstable_by(|a, b| a.name.cmp(&b.name));
        profiles
    }
}

#[cfg(test)]
mod tests {
    use std::path::Path;

    use crate::config::Config;

    #[test]
    fn will_load_default_if_config_does_not_exist() {
        assert!(!Config::from_path(Path::new("/non/existing/file.json")).debug_view_enabled,)
    }
}
